/* cppblas.cpp -- C++ interface to C-BLAS functions

    Copyright (C) 2015 European Centre for Medium-Range Weather Forecasts

    Author: Robin Hogan <r.j.hogan@ecmwf.int>

    This file is part of the Adept library.

   This file provides a C++ interface to selected Level-2 and -3 BLAS
   functions in which the precision of the arguments (float versus
   double) is inferred via overloading

*/

#include <adept/cppblas.h>

#ifdef HAVE_CONFIG_H
// Get C-BLAS header file name from config.h
#include "config.h"
#ifdef CBLAS_HEADER
#include CBLAS_HEADER
// Note that the following functions are only defined if C-BLAS is
// present

namespace adept {

  namespace internal {
    
    // Matrix-matrix multiplication for general dense matrices
#define ADEPT_DEFINE_GEMM(T, FUNC, FUNC_COMPLEX)			\
    void cppblas_gemm(const enum CBLAS_ORDER Order,			\
		      const enum CBLAS_TRANSPOSE TransA,		\
		      const enum CBLAS_TRANSPOSE TransB,		\
		      const int M, const int N,				\
		      const int K, const T alpha, const T *A,		\
		      const int lda, const T *B, const int ldb,		\
		      const T beta, T *C, const int ldc) {		\
      FUNC(static_cast<typename ::CBLAS_ORDER>(Order),			\
	   static_cast<typename ::CBLAS_TRANSPOSE>(TransA),		\
	   static_cast<typename ::CBLAS_TRANSPOSE>(TransB),		\
	   M, N, K, alpha, A, lda,					\
	   B, ldb, beta, C, ldc);					\
    }
    ADEPT_DEFINE_GEMM(double, cblas_dgemm, cblas_zgemm);
    ADEPT_DEFINE_GEMM(float,  cblas_sgemm, cblas_cgemm);
#undef ADEPT_DEFINE_GEMM
    
    // Matrix-vector multiplication for a general dense matrix
#define ADEPT_DEFINE_GEMV(T, FUNC, FUNC_COMPLEX)			\
    void cppblas_gemv(const enum CBLAS_ORDER Order,			\
		      const enum CBLAS_TRANSPOSE TransA,		\
		      const int M, const int N,				\
		      const T alpha, const T *A, const int lda,		\
		      const T *X, const int incX, const T beta,		\
		      T *Y, const int incY) {				\
      FUNC(static_cast<typename ::CBLAS_ORDER>(Order),			\
	   static_cast<typename ::CBLAS_TRANSPOSE>(TransA),		\
	   M, N, alpha, A, lda, X, incX, beta, Y, incY);	\
    }
    ADEPT_DEFINE_GEMV(double, cblas_dgemv, cblas_zgemv);
    ADEPT_DEFINE_GEMV(float,  cblas_sgemv, cblas_cgemv);
#undef ADEPT_DEFINE_GEMV
    
    // Matrix-matrix multiplication where matrix A is symmetric
#define ADEPT_DEFINE_SYMM(T, FUNC, FUNC_COMPLEX)			\
    void cppblas_symm(const enum CBLAS_ORDER Order,			\
		      const enum CBLAS_SIDE Side,			\
		      const enum CBLAS_UPLO Uplo,			\
		      const int M, const int N,				\
		      const T alpha, const T *A, const int lda,		\
		      const T *B, const int ldb, const T beta,		\
		      T *C, const int ldc) {				\
      FUNC(static_cast<typename ::CBLAS_ORDER>(Order),			\
	   static_cast<typename ::CBLAS_SIDE>(Side),			\
	   static_cast<typename ::CBLAS_UPLO>(Uplo),			\
	   M, N, alpha, A, lda, B, ldb, beta, C, ldc);			\
    }
    ADEPT_DEFINE_SYMM(double, cblas_dsymm, cblas_zsymm);
    ADEPT_DEFINE_SYMM(float,  cblas_ssymm, cblas_csymm);
#undef ADEPT_DEFINE_SYMM
    
    // Matrix-vector multiplication where the matrix is symmetric
#define ADEPT_DEFINE_SYMV(T, FUNC, FUNC_COMPLEX)			\
    void cppblas_symv(const enum CBLAS_ORDER Order,			\
		      const enum CBLAS_UPLO Uplo,			\
		      const int N, const T alpha, const T *A,		\
		      const int lda, const T *X, const int incX,	\
		      const T beta, T *Y, const int incY) {		\
      FUNC(static_cast<typename ::CBLAS_ORDER>(Order),			\
	   static_cast<typename ::CBLAS_UPLO>(Uplo),			\
	   N, alpha, A, lda, X, incX, beta, Y, incY);			\
    }					
    ADEPT_DEFINE_SYMV(double, cblas_dsymv, cblas_zsymv);
    ADEPT_DEFINE_SYMV(float,  cblas_ssymv, cblas_csymv);
#undef ADEPT_DEFINE_SYMV
    
    // Matrix-vector multiplication for a general band matrix
#define ADEPT_DEFINE_GBMV(T, FUNC, FUNC_COMPLEX)			\
    void cppblas_gbmv(const enum CBLAS_ORDER Order,			\
		      const enum CBLAS_TRANSPOSE TransA,		\
		      const int M, const int N,				\
		      const int KL, const int KU, const T alpha,	\
		      const T *A, const int lda, const T *X,		\
		      const int incX, const T beta, T *Y,		\
		      const int incY) {					\
      FUNC(static_cast<typename ::CBLAS_ORDER>(Order),			\
	   static_cast<typename ::CBLAS_TRANSPOSE>(TransA),		\
	   M, N, KL, KU, alpha, A, lda, X, incX,			\
	   beta, Y, incY);						\
    }
    ADEPT_DEFINE_GBMV(double, cblas_dgbmv, cblas_zgbmv);
    ADEPT_DEFINE_GBMV(float,  cblas_sgbmv, cblas_cgbmv);
#undef ADEPT_DEFINE_GBMV
  
  } // End namespace internal
  
} // End namespace adept
  
  
#endif
#endif
