/* cpplapack.h -- C++ interface to LAPACK

    Copyright (C) 2015 European Centre for Medium-Range Weather Forecasts

    Author: Robin Hogan <r.j.hogan@ecmwf.int>

    This file is part of the Adept library.
*/

#ifndef AdeptCppLapack_H
#define AdeptCppLapack_H 1                       

#ifdef HAVE_CONFIG_H
// Get C-BLAS header file name from config.h
#include "config.h"
#ifdef HAVE_LAPACKE
#include <lapacke.h>
// Note that the following functions are only defined if LAPACKE is
// present


namespace adept {

  // Overloaded functions serve two purposes: provides both single &
  // double precision versions, and prevents the huge lapacke.h having
  // to be included in all user code
  namespace internal {
    // Factorize a general matrix
    inline
    lapack_int cpplapack_getrf(lapack_int n, float* a,  lapack_int lda,
			       lapack_int* ipiv) {
      return LAPACKE_sgetrf(LAPACK_COL_MAJOR, n, n, a, lda, ipiv);
    }
    inline
    lapack_int cpplapack_getrf(lapack_int n, double* a, lapack_int lda,
			       lapack_int* ipiv) {
      return LAPACKE_dgetrf(LAPACK_COL_MAJOR, n, n, a, lda, ipiv);
    }

    // Invert a general matrix
    inline
    lapack_int cpplapack_getri(lapack_int n, float* a,  lapack_int lda,
			       const lapack_int* ipiv) {
      return LAPACKE_sgetri(LAPACK_COL_MAJOR, n, a, lda, ipiv);
    }
    inline
    lapack_int cpplapack_getri(lapack_int n, double* a, lapack_int lda,
			       const lapack_int* ipiv) {
      return LAPACKE_dgetri(LAPACK_COL_MAJOR, n, a, lda, ipiv);
    }

    // Factorize a symmetric matrix
    inline
    lapack_int cpplapack_sytrf(char uplo, lapack_int n, float* a,
			       lapack_int lda, lapack_int* ipiv) {
      return LAPACKE_ssytrf(LAPACK_COL_MAJOR, uplo, n, a, lda, ipiv);
    }
    inline
    lapack_int cpplapack_sytrf(char uplo, lapack_int n, double* a, 
			       lapack_int lda, lapack_int* ipiv) {
      return LAPACKE_dsytrf(LAPACK_COL_MAJOR, uplo, n, a, lda, ipiv);
    }

    // Invert a symmetric matrix
    inline
    lapack_int cpplapack_sytri(char uplo, lapack_int n, float* a,
			       lapack_int lda, const lapack_int* ipiv) {
      return LAPACKE_ssytri(LAPACK_COL_MAJOR, uplo, n, a, lda, ipiv);
    }
    inline
    lapack_int cpplapack_sytri(char uplo, lapack_int n, double* a,
			       lapack_int lda, const lapack_int* ipiv) {
      return LAPACKE_dsytri(LAPACK_COL_MAJOR, uplo, n, a, lda, ipiv);
    }

    // Solve system of linear equations with general matrix
    inline
    lapack_int cpplapack_gesv(lapack_int n, lapack_int nrhs, float* a,
			      lapack_int lda,
			      lapack_int* ipiv, float* b, lapack_int ldb) {
      return LAPACKE_sgesv(LAPACK_COL_MAJOR, n, nrhs, a, lda, ipiv, b, ldb);
    }
    inline
    lapack_int cpplapack_gesv(lapack_int n, lapack_int nrhs, double* a,
			      lapack_int lda,
			      lapack_int* ipiv, double* b, lapack_int ldb) {
      return LAPACKE_dgesv(LAPACK_COL_MAJOR, n, nrhs, a, lda, ipiv, b, ldb);
    }

    // Solve system of linear equations with symmetric matrix
    inline
    lapack_int cpplapack_sysv(char uplo, lapack_int n, lapack_int nrhs,
			      float* a, lapack_int lda, lapack_int* ipiv,
			      float* b, lapack_int ldb) {
      return LAPACKE_ssysv(LAPACK_COL_MAJOR, uplo, n, nrhs, a, lda, ipiv, b, 
			   ldb);
    }
    inline
    lapack_int cpplapack_sysv(char uplo, lapack_int n, lapack_int nrhs,
			      double* a, lapack_int lda, lapack_int* ipiv, 
			      double* b, lapack_int ldb) {
      return LAPACKE_dsysv(LAPACK_COL_MAJOR, uplo, n, nrhs, a, lda, ipiv, b,
			   ldb);
    }

  }
}

#endif
#endif

#endif
